import os
import math
import matplotlib.pyplot as plt
from PIL import Image


def plot_path(result):
    plotted = []
    ax = None
    fix = None

    for a in result.plan.actions:
        if a.motion_paths is not None:
            for key in a.motion_paths:
                # print(f"  {key} -> {a.motion_paths[key]}")
                # print(f"  Map: {key.starting.type.occupancy_map}")
                map_file = key.starting.type.occupancy_map.filename
                map_path = map_file.replace(map_file.split(os.sep)[-1], "")
                path = a.motion_paths[key]
                image_file = None
                resolution = 1.0
                f = open(map_file)
                for l in f.readlines():
                    if "image: " in l:
                        image_file = map_path + os.sep + l.replace("image: ", "").strip()
                    if "resolution: " in l:
                        resolution = float(l.replace("resolution: ", "").strip())
                if image_file is not None:
                    if image_file not in plotted:
                        im = Image.open(image_file)
                        fig, ax = plt.subplots()
                        ax.imshow(im)
                        plotted.append(image_file)

                    for step in path:
                        x = step[0][0]/resolution
                        y = step[0][1]/resolution
                        w = step[0][2]
                        x_d = math.cos(w)
                        y_d = math.sin(w)
                        plt.arrow(x, im.height - y, x_d, y_d)
    plt.show()



import unittest

from unified_planning.test.examples import get_example_problems
from unified_planning.engines import PlanGenerationResultStatus, CompilationKind
from unified_planning.engines.results import POSITIVE_OUTCOMES

from up_spiderplan.solver import EngineImpl

problems = get_example_problems()


basic_nc = problems["basic_nested_conjunctions"].problem
robot = problems["robot"].problem
basic_wc = problems["basic_with_costs"].problem
counter_2_50 = problems["counter_to_50"].problem
match_cellar = problems["matchcellar"].problem

robot_no_neg = problems["robot_no_negative_preconditions"].problem


class TestClassical(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        self.problems = get_example_problems()
        self.planner = EngineImpl()

    def test_basic(self):
        problem = self.problems["basic"].problem

        final_report = self.planner.solve(problem)
        a = problem.action("a")

        print(final_report)
    
        plan = final_report.plan
        self.assertEqual(
            final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING
        )
        self.assertEqual(len(plan.actions), 1)
        self.assertEqual(plan.actions[0].action, a)
        self.assertEqual(len(plan.actions[0].actual_parameters), 0)


    def test_robot_no_negative_preconditions(self):
        problem = self.problems["robot_no_negative_preconditions"].problem
        print(problem)
        final_report = self.planner.solve(problem)
        #a = problem.action("a")

        print(final_report)
    
        plan = final_report.plan
        self.assertEqual(
            final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING
        )
        self.assertEqual(len(plan.actions), 1)
        # self.assertEqual(plan.actions[0].action, a)
        self.assertEqual(len(plan.actions[0].actual_parameters), 2)
        
        
    def test_basic_2(self):
        from planning_tests.classical_planning.problems.problem_basic import UPBasic
        problem = UPBasic(1).get_problem()
        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)

    def test_basic_unsolvable(self):
        from planning_tests.classical_planning.problems.problem_basic_unsolvable import UPBasicUnsolvable
        problem = UPBasicUnsolvable(1).get_problem()
        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.UNSOLVABLE_PROVEN)

    def test_depot_1(self):
        from planning_tests.classical_planning.pddl_problems.depots.depots import depots_pfile1
        problem = depots_pfile1(1).get_problem()

        print(problem)

        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)

    def test_depot_2(self):
        from planning_tests.classical_planning.pddl_problems.depots.depots import depots_pfile2
        problem = depots_pfile2(1).get_problem()

        print(problem)

        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)

    def test_depot_3(self):
        from planning_tests.classical_planning.pddl_problems.depots.depots import depots_pfile3
        problem = depots_pfile3(1).get_problem()
        print(problem)

        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)

    def test_tpp_1(self):
        from planning_tests.classical_planning.pddl_problems.tpp.tpp import tpp_pfile1
        problem = tpp_pfile1(1).get_problem()

        print(problem)

        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)

    def test_tpp_2(self):
        from planning_tests.classical_planning.pddl_problems.tpp.tpp import tpp_pfile2
        problem = tpp_pfile2(1).get_problem()

        print(problem)

        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)

    def test_tpp_3(self):
        from planning_tests.classical_planning.pddl_problems.tpp.tpp import tpp_pfile3
        problem = tpp_pfile3(1).get_problem()

        print(problem)

        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)

    def test_tpp_6(self):
        from planning_tests.classical_planning.pddl_problems.tpp.tpp import tpp_pfile6
        problem = tpp_pfile6(1).get_problem()

        print(problem)

        result = self.planner.solve(problem)
        self.assertNotEqual(result.status, PlanGenerationResultStatus.INTERNAL_ERROR)
        # result status must reflect that there is no plan
        self.assertEqual(result.status, PlanGenerationResultStatus.SOLVED_SATISFICING)



import unified_planning.engines.results as results

from unified_planning.test.examples import tamp

from combined_task_motion_planning.office.test_01 import TampOffice01
from combined_task_motion_planning.office.test_02 import TampOffice02
from combined_task_motion_planning.office.test_03 import TampOffice03

from combined_task_motion_planning.construction.test_01 import TampConstruction01
from combined_task_motion_planning.construction.test_02 import TampConstruction02
from combined_task_motion_planning.construction.test_03 import TampConstruction03

from combined_task_motion_planning.logistics.test_01 import TampLogistics01

from solver import EngineImpl

from util import plot_path

selected = TampOffice03(1).get_problem()

print(selected)

solver = EngineImpl(run_docker=True)
result = solver.solve(selected)

if result.status in results.POSITIVE_OUTCOMES:
    for a in result.plan.actions:
        print(a)
        print(a.motion_paths)

    plot_path(result)
else:
    print("NO-PLAN-FOUND")



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import subprocess
import shutil
import os
import time

from functools import partial
from typing import IO, Callable, List, Dict, Optional, Set, Tuple
import warnings
import unified_planning as up
import unified_planning.engines
from aiddl_core.container import Container, Entry
from unified_planning.exceptions import UPUnsupportedProblemTypeError
from unified_planning.engines import PlanGenerationResultStatus, Credits
from unified_planning.model import FNode, ProblemKind, Type as UPType
from unified_planning.model.tamp import ReedsSheppPath
from unified_planning.model.tamp import Waypoints
from aiddl_core.representation import Term, Sym, Var
from aiddl_core.representation import Tuple as AiddlTuple

from aiddl_core.util.logger import Logger
from aiddl_external_grpc_python.container import GrpcFunction

from up_spiderplan.converter import UpCdbConverter

import tempfile

credits = Credits('spiderplan',
                  'Örebro University',
                  'Uwe Köckemann',
                  'spiderplan.org',
                  'GPL3 LICENSE',
                  'Spiderplan is an extentable constraint-based hybrid planner.',
                  'Spiderplan is an extentable constraint-based hybrid planner.'
                )

SPIDER_dst = tempfile.gettempdir() + "/"
SPIDER_PUBLIC = "up-spiderplan"
SPIDER_TAG = "master"
SPIDER_REPO = "https://github.com/aiplan4eu/up-spiderplan.git"
SPIDER_PORT = 8061

class EngineImpl(
    unified_planning.engines.Engine,
    unified_planning.engines.mixins.OneshotPlannerMixin):
    def __init__(self, run_docker=True, **options):
        self._run_docker = run_docker
        self._problem = None
        self.conv = UpCdbConverter()
        if len(options) > 0:
            raise

    def copy_file_to_container(self, path, name):
        cmd = f'docker cp "{path}/{name}" up-spiderplan-server:/planner/{name}'
        print("CMD:", cmd)
        os.system(cmd)

    def install_grpc_server(self):
        subprocess.run(["git", "clone", "-b", SPIDER_TAG, SPIDER_REPO])
        shutil.move(SPIDER_PUBLIC, SPIDER_dst)
        curr_dir = os.getcwd()
        os.chdir(SPIDER_dst  + SPIDER_PUBLIC + "/spiderplan-grpc-server")
        os.system("docker compose build")
        os.chdir(curr_dir)

    def grpc_server_installed(self):
        return os.path.exists(SPIDER_dst  + SPIDER_PUBLIC)

    def start_grpc_server(self):
        curr_dir = os.getcwd()
        os.chdir(SPIDER_dst  + SPIDER_PUBLIC + "/spiderplan-grpc-server")
        os.system("docker compose up -d")
        os.chdir(curr_dir)


    def stop_grpc_server(self):
        os.system(f"docker cp up-spiderplan-server:/planner/search.dot ./search.dot")
        os.system(f"docker cp up-spiderplan-server:/planner/stopwatch.txt ./stopwatch.txt")

        curr_dir = os.getcwd()
        os.chdir(SPIDER_dst  + SPIDER_PUBLIC + "/spiderplan-grpc-server")
        os.system("docker compose stop")
        os.chdir(curr_dir) 


    @property
    def name(self) -> str:
        return "Spiderplan"

    def solve(self, problem: 'up.model.Problem',
                callback: Optional[Callable[['up.engines.PlanGenerationResult'], None]] = None,
                timeout: Optional[float] = None,
                output_stream: Optional[IO[str]] = None) -> 'up.engines.results.PlanGenerationResult':
        '''This function converts a UP problem to a constraint database and uses SpiderPlan to resolve any open flaws.'''
        if not self.supports(problem.kind):
            print('SpiderPlan cannot solve this kind of problem!')
            # raise UPUnsupportedProblemTypeError('Spiderplan cannot solve this kind of problem!')

        if timeout is not None:
            warnings.warn('SpiderPlan does not support timeout.', UserWarning)
        if output_stream is not None:
            warnings.warn('SpiderPlan does not support output stream.', UserWarning)

        if self._run_docker and not self.grpc_server_installed():
            self.install_grpc_server()

        self._problem = problem
        cdb = self._convert(problem)

        # print(Logger.pretty_print(cdb, 1))

        solution_cdb = self._solve(cdb)

        if solution_cdb == Sym("NIL"):
            result = up.engines.PlanGenerationResult(PlanGenerationResultStatus.UNSOLVABLE_PROVEN, None, self.name)
            result.metrics = {"engine_internal_time": self.internal_engine_time}
            return result

        plan = self._extract_plan(solution_cdb)
        result = up.engines.PlanGenerationResult(PlanGenerationResultStatus.SOLVED_SATISFICING, plan, self.name)
        result.metrics = {"engine_internal_time": self.internal_engine_time}
        return result

    def _convert(self, problem: 'unified_planning.model.Problem') -> Term:
        return self.conv(problem)

    def _solve(self, cdb: 'aiddl_core.representation.Set') -> Term:
        # Call Spiderplan and get resulting CDB or NIL if no solution exists.
        container = Container()
        spiderplan_proxy = GrpcFunction("localhost", SPIDER_PORT, Sym("org.spiderplan.unified-planning.basic-graph-search"), container)

        mod_name = Sym("aiplan4eu.up-spiderplan.up-input")
        container.add_module(mod_name)
        e = Entry(Sym("org.aiddl.type.term"), Sym("problem"), cdb)
        container.set_entry(e, mod_name)
        container.export(mod_name, "converted-from-up.aiddl")

        # print("SpiderPlan Problem:")
        # print(Logger.pretty_print(cdb, 0))

        if self._run_docker:
            self.start_grpc_server()
        success = False
        while not success:
            from grpc._channel import _InactiveRpcError
            try:
                answer = spiderplan_proxy(cdb)
                self._extract_internal_engine_time()
                success = True
            except _InactiveRpcError as e:
                # print("Waiting for Spiderplan gRPC server...")
                #print(e)
                success = False
                time.sleep(0.1)

        if self._run_docker:
            self.stop_grpc_server()

        # print("SpiderPlan Solution:")
        # print(Logger.pretty_print(answer, 0))

        return answer

    def _extract_internal_engine_time(self):
        f = open("stopwatch.txt", "r")
        self.internal_engine_time = None
        for line in f.readlines():
            if line.startswith("[SpiderPlan] Main: "):
                self.internal_engine_time = line.split()[2]
        f.close()

    def _extract_plan(self, cdb: 'aiddl_core.representation.Set') -> up.plans.Plan:
        # Convert CDB to a plan type supported by UP
        # -> sequence or partial-order based on EST in propagated STN
        op_map = {}

        # print(Logger.pretty_print(cdb, 0))

        for a in cdb.get((Sym("operator"))):
            op_map[a[Sym("name")]] = a

        path_expressions = {}

        for mc in cdb.get_or_default(Sym("motion"), []):
            if mc[0] == Sym("path") and not isinstance(mc[1], Var):
                if mc[1] not in path_expressions.keys():
                    path_expressions[mc[1]] = []
                path_expressions[mc[1]].append(mc)

        instance_map = {}
        plan_stmts = []
        for s in cdb[Sym("statement")]:
            variable = s[1].key
            if variable in op_map.keys():
                earliest_start_time = cdb[Sym("propagated-value")][AiddlTuple(Sym("ST"), s[0])][0]
                plan_stmts.append((earliest_start_time, s))

                operator = op_map[variable]

                # print(earliest_start_time, ":", s)
                # print(Logger.pretty_print(operator, 1))

                id_var = operator[Sym("id")]
                interval_pattern = operator[Sym("interval")]
                sub = interval_pattern.match(s[0])
                id = sub.substitute(id_var)

                instance_map[s] = id

                if a.get(Sym("constraints")).contains_key(Sym("motion")):
                    mcs = a.get(Sym("constraints")).get(Sym("motion"))




        plan_stmts = sorted(plan_stmts, key=lambda x: x[0])

        actions: List[up.plans.ActionInstance] = []
        # print("Plan statements:")
        for pair in cdb[Sym("goal.plan.latest")]:

            # print(plan_stmts)
            # print(pair)

            s = pair[1]
            # print(s)
            id = instance_map[s]

            aiddl_action = s[1].key
            action_params = []
            if isinstance(aiddl_action, AiddlTuple):
                action_name = str(aiddl_action[0])
                for p in range(1, len(aiddl_action)):
                    action_params.append(str(aiddl_action[p]))
            else:
                action_name = str(aiddl_action)

            paths = {}
            if id in path_expressions.keys():
                for pathExp in path_expressions[id]:
                    # print(pathExp)
                    path = pathExp[6].unpack()
                    upPath = ReedsSheppPath(path)

                    mover = str(pathExp[2])
                    start = str(pathExp[3])
                    wps = []
                    if isinstance(pathExp[4], List):
                        for wp in pathExp[4]:
                            wps.append(self._problem.object(str(wp)))
                    else:
                        wps.append(self._problem.object(str(pathExp[4])))

                    upMotionCon = Waypoints(
                        self._problem.object(mover),
                        self._problem.object(start),
                        wps)
                    paths[upMotionCon] = path

            up_action = self._problem.action(action_name)
            expr_manager = self._problem.environment.expression_manager
            param = tuple(expr_manager.ObjectExp(self._problem.object(o_name)) for o_name in action_params)
            if len(paths) == 0:
                actions.append(up.plans.ActionInstance(up_action, param))
            else:
                actions.append(up.plans.ActionInstance(up_action, param, None, motion_paths=paths))

        # for up_action in actions:
            # print(up_action)
        #    for mc in up_action.motion_paths.keys():
                # print(mc, up_action.motion_paths[mc])

        return up.plans.SequentialPlan(actions)


    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class('ACTION_BASED')
        supported_kind.set_problem_class('TAMP')
        supported_kind.set_typing('FLAT_TYPING')
        supported_kind.set_typing('HIERARCHICAL_TYPING')
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        return supported_kind

    @staticmethod
    def supports(problem_kind: 'up.model.ProblemKind') -> bool:
        return problem_kind <= EngineImpl.supported_kind()

    @staticmethod
    def is_oneshot_planner():
        return True

    @staticmethod
    def is_grounder():
        return False

    @staticmethod
    def get_credits(**kwargs) -> Optional[unified_planning.engines.Credits]:
        return credits

    def destroy(self):
        pass



import os

import unified_planning.model.problem
from aiddl_core.util.logger import Logger
from unified_planning.model.tamp.objects import MotionModels
from unified_planning.model import InstantaneousAction, StartTiming, TimepointKind, Timing, TimeInterval
from unified_planning.model.tamp import InstantaneousMotionAction, Waypoints, ConfigurationObject
from unified_planning.shortcuts import ConfigurationType
from unified_planning.model.operators import OperatorKind
from unified_planning.model.problem import Problem

from aiddl_core.representation import Sym, Num, Real, Substitution, Str
from aiddl_core.representation import Boolean
from aiddl_core.representation import Tuple
from aiddl_core.representation import List
from aiddl_core.representation import Set
from aiddl_core.representation import Int
from aiddl_core.representation import KeyVal
from aiddl_core.representation import Var
from aiddl_core.representation import Inf
from aiddl_core.parser.parser import parse_term

from aiddl_core.util.logger import Logger

OpMap = {}
OpMap[OperatorKind.AND] = Sym("org.aiddl.eval.logic.and")
OpMap[OperatorKind.OR] = Sym("org.aiddl.eval.logic.or")
OpMap[OperatorKind.NOT] = Sym("org.aiddl.eval.logic.not")
OpMap[OperatorKind.IMPLIES] = Sym("org.aiddl.eval.logic.implies")
OpMap[OperatorKind.IFF] = Sym("org.aiddl.eval.logic.iff")
OpMap[OperatorKind.EXISTS] = Sym("org.aiddl.eval.logic.exists")
OpMap[OperatorKind.FORALL] = Sym("org.aiddl.eval.logic.forall")
OpMap[OperatorKind.PLUS] = Sym("org.aiddl.eval.numerical.add")
OpMap[OperatorKind.MINUS] = Sym("org.aiddl.eval.numerical.sub")
OpMap[OperatorKind.TIMES] = Sym("org.aiddl.eval.numerical.mult")
OpMap[OperatorKind.DIV] = Sym("org.aiddl.eval.numerical.div")
OpMap[OperatorKind.EQUALS] = Sym("org.aiddl.eval.equals")
OpMap[OperatorKind.LE] = Sym("org.aiddl.eval.numerical.less-than-eq")
OpMap[OperatorKind.LT] = Sym("org.aiddl.eval.numerical.less-than")


def merge_cdb(a, b):
    result = {}
    for kvp in a:
        k_a = kvp.key
        if b.contains_key(k_a):
            result[k_a] = kvp.value.add_all(b[k_a])
        else:
            result[k_a] = kvp.value

    for kvp in b:
        k_b = kvp.key
        if k_b not in result.keys():
            result[k_b] = kvp.value

    r = []
    for k in result.keys():
        r.append(KeyVal(k, result[k]))
    return Set(set(r)) #  for k, v in result])
        
        
class UpCdbConverter:

    def __init__(self):
        self.next_id = 0
        self.map_look_up = {}
        self.map_files = {}

    def __call__(self, problem):
        cdb = Set([])
        init = self._convert_initial_values(problem.initial_values)
        # print(init)
        cdb = merge_cdb(cdb, init)
        goal = self._convert_goal(problem.goals)
        cdb = merge_cdb(cdb, goal)

        cdb = merge_cdb(cdb, self._convert_motion_info(problem))

        type_look_up = {}
        self._convert_user_types(problem, type_look_up)

        operators = self._convert_operators(problem.actions, type_look_up)
        cdb = merge_cdb(cdb, operators)

        signatures = self._convert_fluents(problem, type_look_up)
        cdb = merge_cdb(cdb, signatures)

        cdb = merge_cdb(cdb, self._convert_domains(type_look_up))

        return cdb
        
    def _convert_fnode(self, n):
        # print("=== FNODE: %s (%s) ===" % (str(n), str(type(n))))
        term = None

        if n.is_fluent_exp():
            if len(n.args) == 0:
                term = Sym(str(n))
            else:
                args = [Sym(n.fluent().name)]
                for x in n.args:
                    args.append(self._convert_fnode(x))

                term = Tuple(args)
                
        elif n.is_bool_constant():
            term = Boolean(n.bool_constant_value())
        elif n.is_int_constant():
            term = Int(n.int_constant_value())
        elif n.is_object_exp():
            term = Sym(str(n))
        elif n.is_parameter_exp():
            term = Var(str(n))

        if term is None:
            raise ValueError("Fluent not supported: %s" % str(n))

        # print("=== TERM: %s ===" % (str(term)))
        return term

    def _convert_fluents(self, problem: Problem, t_look_up):
        signatures = []
        for f in problem.fluents:
            name = Sym(str(f.name))
            r_type = self._convert_type(f.type, t_look_up)
            signature = [self._convert_type(s.type, t_look_up) for s in f.signature]
            if len(signature) == 0:
                signatures.append(KeyVal(name, r_type))
            else:
                # print(f)
                signatures.append(KeyVal(Tuple([name] + signature), r_type))

        return Set([
            KeyVal(Sym("signature"), Set(signatures))
        ])

    def _convert_domains(self, t_look_up):
        domains = []
        for d in t_look_up.values():
            domains.append(d)
        return Set([KeyVal(Sym("domain"), Set(domains))])

    def _convert_object(self, o):
        return Sym(o.name)

    def _convert_constraint(self, c, fluents):
        # print("Converting constraint:", c)
        if c.node_type in OpMap.keys():
            exp = [OpMap[c.node_type]]
            for arg in c.args:
                exp.append(self._convert_constraint(arg, fluents))
            term = Tuple(exp)
        else:
            
            if c.is_fluent_exp():
                if c in fluents.keys():
                    term = fluents[c]
                else:
                    self.next_id += 1
                    term = Var("x%d" % self.next_id)
                    fluents[c] = term
            else:
                term = self._convert_fnode(c)
                if isinstance(term, Var):
                    fluents[c] = term
                
        return term

    def _convert_condition(self, c, is_goal=True, op_id=None, effect_vars=[]):
        # UP conditions may correspond to various constraint types
        # Fluent -> goal/precondition + temporal
        # (not Fluent) -> statement with false value + temporal
        # boolean or arithmetic formula -> AIDDL eval constraint expression
        # -> need to track fluents and assign variables + preconditions when they appear (e.g., (and a (or b c)) needs SVAs for a, b, and c
        # -> try to propagate (planning easier if we know which value we need, otherwise we may pick random values and reject based on constraints. in the example above, at least a has to be true)
        # -> CAST AS CSP and link fluents to values of SVAs
        # CSP solver can propagate if single value possible. Otherwise search over options. This nicely works with conjunctive goals
        
        statements = []
        temporal = []
        goals = []
        preconditions = []
        csp = []
        
        if c.is_fluent_exp() or (c.is_not() and c.args[0].is_fluent_exp()):
            self.next_id += 1
            if is_goal:
                i = Sym("G%d" % self.next_id)
            else:
                i = Tuple([Sym("P%d" % self.next_id), Var("ID")])
            if c.is_not():
                g_var = self._convert_fnode(c.args[0])
            else:
                g_var = self._convert_fnode(c)
            g_val = Boolean(not c.is_not())
            
            s = Tuple([i, KeyVal(g_var, g_val)])
            d = Tuple([Sym("duration"), i, Tuple([Int(1), Inf.pos()])])
            if is_goal:
                goals.append(s)
            else:
                preconditions.append(s)
            temporal.append(d)
            if op_id is not None and g_var in effect_vars:
                temporal.append(Tuple(Sym("before"), i, op_id, Tuple(Int(1), Int(1))))
            elif op_id is not None:
                temporal.append(Tuple(Sym("contains"), i, op_id, Tuple(Int(1), Inf.pos()),Tuple(Int(1), Inf.pos())))

        else:
            fluents = {}
            constraint_term = self._convert_constraint(c, fluents)
            scope = []
            for fluent in fluents.keys():
                self.next_id += 1
                if is_goal:
                    i = Sym("G%d" % self.next_id)
                else:
                    i = Tuple([Sym("P%d" % self.next_id), Var("ID")])
                var = self._convert_fnode(fluent)
                val = fluents[fluent]
                if var != val:
                    s = Tuple([i, KeyVal(var, val)])
                    d = Tuple([Sym("duration"), i, Tuple([Int(1), Inf.pos()])])
                    if is_goal:
                        goals.append(s)
                    else:
                        preconditions.append(s)
                    temporal.append(d)
                    if op_id is not None and var in effect_vars:
                        temporal.append(Tuple(Sym("before"), i, op_id, Tuple(Int(1),  Int(1))))
                    elif op_id is not None:
                        temporal.append(Tuple(Sym("contains"), i, op_id, Tuple(Int(1), Inf.pos()),Tuple(Int(1), Inf.pos())))
                scope.append(fluents[fluent])

            vars = List(scope)
            domains = List([KeyVal(x,
                                     List([Boolean(True),
                                           Boolean(False)]))
                            for x in vars])
            scope = Tuple(scope)
            constraint_exp = [
                KeyVal(Sym("variables"), vars),
                KeyVal(Sym("domains"), domains),
                KeyVal(Sym("constraints"),
                         Set([Tuple([
                             scope,
                             Tuple([
                                 Sym("org.aiddl.eval.lambda"),
                                 scope,
                                 constraint_term])])]))]
            
            
            csp += constraint_exp

        cdb = []
        if len(goals) > 0:
            cdb.append(KeyVal(Sym("goal"), List(goals)))
        if len(statements) > 0:
            cdb.append(KeyVal(Sym("statement"), List(statements)))
        if len(temporal) > 0:
            cdb.append(KeyVal(Sym("temporal"), List(temporal)))
        if len(preconditions) > 0:
            cdb.append(KeyVal(Sym("preconditions"), List(preconditions)))
        if len(csp) > 0:
            cdb.append(KeyVal(Sym("csp"), List(csp)))
            
        return Set(cdb)

    def _convert_effect(self, e, op_id, effect_id, durative=False):
        statement = []
        temporal = []
        interval = Tuple([Sym("E%d" % effect_id), Var('ID')])
        s = Tuple([interval,
                   KeyVal(
                       self._convert_fnode(e.fluent),
                       self._convert_fnode(e.value))])

        d = Tuple([Sym("duration"), interval, Tuple([Int(1), Inf.pos()])])
        m = Tuple([Sym("before"), op_id, interval, Tuple(Int(1),  Int(1))])

        statement.append(s)
        if not durative:
            temporal.append(m)
            m = Tuple(Sym("before"), op_id, interval, Tuple(Int(1),  Int(1)))
        temporal.append(d)
        return Set([
            KeyVal(Sym("effects"), List(statement)),
            KeyVal(Sym("temporal"), List(temporal))])
        
    def _convert_initial_values(self, init_vals):
        statements = []
        temporal = []
        for e in init_vals:
            self.next_id += 1
            interval = Sym("s%d" % self.next_id)
            s = Tuple([interval,
                       KeyVal(
                           self._convert_fnode(e),
                           self._convert_fnode(init_vals[e]))])

            r = Tuple([Sym("release"), interval, Tuple([Int(0), Int(0)])])
            d = Tuple([Sym("duration"), interval, Tuple([Int(1), Inf.pos()])])
            statements.append(s)
            temporal.append(r)
            temporal.append(d)
        return Set([
            KeyVal(Sym("statement"), List(statements)),
            KeyVal(Sym("temporal"), List(temporal))])

    def _convert_goal(self, goal):
        goal_cdb = Set([])
        for g in goal:
            g_cdb = self._convert_condition(g, is_goal=True)
            goal_cdb = merge_cdb(goal_cdb, g_cdb)
        return goal_cdb

    def _convert_operators(self, ops, t_look_up):
        spider_ops = []
        signatures = []
        for o in ops:
            if isinstance(o, InstantaneousAction):
                spider_op = self._convert_instantaneous_action(o, t_look_up)
            else:
                spider_op = self._convert_durative_action(o, t_look_up)

            # print("SPIDER:", Logger.pretty_print(spider_op, 0))

            name = spider_op[Sym("name")]
            signature = spider_op[Sym("signature")]

            if len(signature) == 0:
                signatures.append(KeyVal(name, Sym("t_bool")))
            else:
                # print(signature)
                sig = [name[0]]
                for p in signature:
                    sig.append(p.value)
                signatures.append(KeyVal(Tuple(sig), Sym("t_bool")))

            spider_ops.append(spider_op)
        return Set([
            KeyVal(Sym("operator"), Set(spider_ops)),
            KeyVal(Sym("signature"), Set(signatures)),
        ])

    def _convert_durative_action(self, a, t_look_up):
        # print("-------------------------------------")
        base_name = Sym(a.name)
        id_var = Var('ID')
        op_id = Tuple([base_name, id_var])
        args = [base_name]
        sig = []
        for p in a.parameters:
            x, t = self._convert_parameter(p, t_look_up)
            args.append(x)
            sig.append(KeyVal(x, t))

        if len(args) == 1:
            name = args[0]
        else:
            name = Tuple(args)
        signature = List(sig)

        preconditions = []
        effects = []
        temporal = [Tuple([Sym("duration"), op_id, Tuple([Int(1), Inf.pos()])])]
        csp = []
        for i, cs in a.conditions.items():
            ac = None
            ti1 = i.lower
            ti2 = i.upper

            if ti1.timepoint.kind == TimepointKind.START and ti2.timepoint.kind == TimepointKind.START:
                delay = ti1.delay
                ac = Tuple(Sym("overlaps"), Var("I"), op_id, Tuple(Int(delay), Inf.pos()))
            elif ti1.timepoint.kind == TimepointKind.START and ti2.timepoint.kind == TimepointKind.END:
                ac = Tuple(Sym("contains"), Var("I"), op_id, Tuple(Int(1), Inf.pos()), Tuple(Int(1), Inf.pos()))

            for c in cs:
                self.next_id += 1
                cond_i = Tuple([Sym("P%d" % self.next_id), Var("ID")])
                sub = Substitution()
                sub.add(Var("I"), cond_i)
                ac_sub = ac.substitute(sub)
                self.next_id -= 1
                cond_aiddl = self._convert_condition(c, is_goal=False, op_id=None)
                if cond_aiddl.contains_key(Sym('preconditions')):
                    for p in cond_aiddl[Sym('preconditions')]:
                        preconditions.append(p)
                if cond_aiddl.contains_key(Sym('temporal')):
                    for p in cond_aiddl[Sym('temporal')]:
                        temporal.append(p)
                if cond_aiddl.contains_key(Sym('csp')):
                    for p in cond_aiddl[Sym('csp')]:
                        csp.append(p)
                # print("COND:", cond_aiddl)
                # print("TEMP:", ac_sub)

                temporal.append(ac_sub)

        for i, es in a.effects.items():
            ac = None
            delay = i.delay
            if i.timepoint.kind == TimepointKind.START:
                ac = Tuple(Sym("started-by"), Var("I"), op_id, Tuple(Int(delay), Int(delay)))
            elif i.timepoint.kind == TimepointKind.END:
                ac = Tuple(Sym("before"), op_id, Var("I"), Tuple(Int(delay), Int(delay)))

            for e in es:
                self.next_id += 1
                effect_id = Tuple([Sym("E%d" % self.next_id), Var("ID")])
                sub = Substitution()
                sub.add(Var("I"), effect_id)
                ac_sub = ac.substitute(sub)
                temporal.append(ac_sub)
                constraints = self._convert_effect(e, op_id, self.next_id, durative=True)
                # print("Effect extracted:", constraints)
                if constraints.contains_key(Sym('effects')):
                    for p in constraints[Sym('effects')]:
                        effects.append(p)
                if constraints.contains_key(Sym('temporal')):
                    for p in constraints[Sym('temporal')]:
                        temporal.append(p)

        constraint_list = []
        constraint_list.append(KeyVal(Sym('temporal'), List(temporal)))
        if len(csp) > 0:
            constraint_list.append(KeyVal(Sym('csp'), List(csp)))

        return Tuple([
            KeyVal(Sym('name'), name),
            KeyVal(Sym('signature'), signature),
            KeyVal(Sym('id'), id_var),
            KeyVal(Sym('interval'), op_id),
            KeyVal(Sym('preconditions'), List(preconditions)),
            KeyVal(Sym('effects'), List(effects)),
            KeyVal(Sym('constraints'), List(constraint_list))
        ])


    
    def _convert_instantaneous_action(self, o, t_look_up):
        base_name = Sym(o.name)
        id_var = Var('ID')
        op_id = Tuple([base_name, id_var])
        args = [base_name]
        sig = []
        for p in o.parameters:
            x, t = self._convert_parameter(p, t_look_up)
            args.append(x)
            sig.append(KeyVal(x, t))

        preconditions = []
        effects = []

        temporal = [Tuple([Sym("duration"), op_id, Tuple([Int(1), Inf.pos()])])]

        csp = []

        effect_vars = []
        effect_id = 0
        for e in o.effects:
            effect_id += 1
            constraints = self._convert_effect(e, op_id, effect_id)
            # print("Effect extracted:", constraints)
            if constraints.contains_key(Sym('effects')):
                for p in constraints[Sym('effects')]:
                    effect_vars.append(p[1].key)
                    effects.append(p)
            if constraints.contains_key(Sym('temporal')):
                for p in constraints[Sym('temporal')]:
                    temporal.append(p)

        for p in o.preconditions:
            constraints = self._convert_condition(p, is_goal=False, op_id=op_id, effect_vars=effect_vars)
            if constraints.contains_key(Sym('preconditions')):
                for p in constraints[Sym('preconditions')]:
                    preconditions.append(p)
            if constraints.contains_key(Sym('temporal')):
                for p in constraints[Sym('temporal')]:
                    temporal.append(p)
            if constraints.contains_key(Sym('csp')):
                for p in constraints[Sym('csp')]:
                    csp.append(p)



        if len(args) == 1:
            name = args[0]
        else:
            name = Tuple(args)
        signature = List(sig)

        # print(preconditions)

        constraint_list = []
        constraint_list.append(KeyVal(Sym('temporal'), List(temporal)))
        if len(csp) > 0:
            constraint_list.append(KeyVal(Sym('csp'), List(csp)))

        if isinstance(o, InstantaneousMotionAction):
            #motion:{
            #    (path ?ID ?r ?l1 ?l2 map-1 ?path)
            #}
            motion_constraints = []
            # print("CONVERTING MOTION CONSTRAINTS")
            for mc in o.motion_constraints:
                if isinstance(mc, Waypoints):
                    movable = self._convert_fnode(mc.movable)
                    start = self._convert_fnode(mc.starting)
                    wps = List([self._convert_fnode(wp) for wp in mc.waypoints])
                    # print(movable)

                    map_name = self.map_look_up[mc.starting.type.occupancy_map]

                    c = Tuple(Sym("path"), Var("ID"), movable, start, wps[0], map_name, Var("path"))
                    # print(c)
                    motion_constraints.append(c)
                else:
                    raise ValueError(f"Unsupported motion constraint of type {type(mc)}: {mc}")

                constraint_list.append(KeyVal(Sym('motion'), List(motion_constraints)))

                
        return Tuple([
            KeyVal(Sym('name'), name),
            KeyVal(Sym('signature'), signature),
            KeyVal(Sym('id'), id_var),
            KeyVal(Sym('interval'), op_id),
            KeyVal(Sym('preconditions'), List(preconditions)),
            KeyVal(Sym('effects'), List(effects)),
            KeyVal(Sym('constraints'), List(constraint_list))
        ])

    def _convert_type(self, t, t_look_up):
        type_name = None
        if t.is_bool_type():
            type_name = Sym("t_bool")
            t_look_up[t] = KeyVal(type_name, List([Boolean(True), Boolean(False)]))
        elif t.is_int_type() or t.is_real_type():
            if t.is_int_type():
                range_type = Sym("int")
            else:
                range_type = Sym("real")
            if t.lower_bound is None:
                lb = Inf.neg()
            else:
                if t.is_int_type():
                    lb = Int(t.lower_bound)
                else:
                    lb = Real(t.lower_bound)
            if t.upper_bound is None:
                ub = Inf.pos()
            else:
                if t.is_int_type():
                    ub = Int(t.upper_bound)
                else:
                    ub = Real(t.upper_bound)
            domain = Tuple([Sym("range"), Tuple([lb, ub]), List([KeyVal(Sym("type"), range_type)])])
            self.next_id += 1
            type_name = Sym(f't_range-{self.next_id}')
            t_look_up[t] = KeyVal(type_name, domain)
        else: # UserType
            type_name = Sym(t.name)

        return type_name

    def _convert_parameter(self, p, t_look_up):
        if p not in t_look_up.keys():
            if p.type.is_bool_type():
                t_look_up[p] = KeyVal(Sym("t_bool"), List([Boolean(True), Boolean(False)]))
            elif p.type.is_int_type() or p.type.is_real_type():
                if p.type.is_int_type():
                    range_type = Sym("int")
                else:
                    range_type = Sym("real")
                if p.type.lower_bound is None:
                    lb = Inf.neg()
                else:
                    lb = Num(p.type.lower_bound)
                if p.type.upper_bound is None:
                    ub = Inf.pos()
                else:
                    ub = Num(p.type.lower_bound)
                domain = Tuple([Sym("range"), Tuple([lb, ub]), List([KeyVal(Sym("type"), range_type)])])
                self.next_id += 1
                t_look_up[p.type] = (Sym(f'range-{self.next_id}'), domain)

                return domain
        
        return (Var(p.name), Sym(p.type.name))

    def _convert_user_types(self, problem, type_look_up):
        for utype in problem.user_types:
            name = Sym(utype.name)
            domain = List([self._convert_object(o) for o in problem.objects(utype)])
            type_look_up[utype] = KeyVal(name, domain)

    def _convert_motion_info(self, problem):
        next_map_idx = 0
        constraints = []
        poses = {}

        # print("GETTING MOTION OBJECT INFO")

        for utype in problem.user_types:
            # print(utype, type(utype))
            if utype.is_configuration_type():
                m = utype.occupancy_map

                if m not in self.map_look_up.keys():
                    next_map_idx += 1
                    name = Sym(f"map-{next_map_idx}")
                    self.map_look_up[m] = name

                    # Copy YAML and image file into running Docker container
                    filename = m.filename.split(os.sep)[-1]
                    path = m.filename.replace(filename, "")
                    image_filename = None
                    f = open(m.filename)
                    for l in f.readlines():
                        if "image" in l:
                            image_filename = l.split("image: ")[1].strip()
                    f.close()
                    self.map_files[name] = m.filename
                        
                    cmd_yaml = f'docker cp "{path}{os.sep}{filename}" up-spiderplan-server:/planner/{filename}'
                    cmd_image = f'docker cp "{path}{os.sep}{image_filename}" up-spiderplan-server:/planner/{image_filename}'
                    os.system(cmd_yaml)
                    os.system(cmd_image)

                    constraints.append(Tuple(Sym("map"), name, Str(filename)))
                else:
                    name = self.map_look_up[m]

                if name not in poses.keys():
                    poses[name] = []

                for o in problem.objects(utype):
                    symbol = Sym(o.name)
                    cfg = Tuple([Real(x) for x in list(o.configuration)])
                    poses[name].append(KeyVal(symbol, cfg))

            elif utype.is_movable_type():
                for o in problem.objects(utype):
                    name = Sym(o.name)
                    footprint = List([Tuple(Real(p[0]), Real(p[1])) for p in o.footprint])
                    parameters = []
                    if o.motion_model == MotionModels.REEDSSHEPP:
                        model = Sym("ReedsSheppCar")
                        parameters.append(KeyVal(Sym("turning-radius"), Real(o.parameters["turning_radius"])))
                    else:
                        raise ValueError(f"Motion model {o.motion_model} is not supported by SpiderPlan")
                    constraints.append(Tuple(Sym("frame"), name, footprint))
                    constraints.append(Tuple(Sym("robot"), name, List(
                        KeyVal(Sym("name"), name),
                        KeyVal(Sym("footprint"), footprint),
                        KeyVal(Sym("model"), model),
                        KeyVal(Sym("turning-radius"), Real(o.parameters["turning_radius"]))
                    )))

        for map_name in poses.keys():
            con = Tuple(Sym("poses"), map_name, List(poses[map_name]))
            constraints.append(con)

        return Set([
            KeyVal(Sym("motion"), Set(constraints))
        ])


